"use strict";
var __extends = (this && this.__extends) || (function () {
    var extendStatics = Object.setPrototypeOf ||
        ({ __proto__: [] } instanceof Array && function (d, b) { d.__proto__ = b; }) ||
        function (d, b) { for (var p in b) if (b.hasOwnProperty(p)) d[p] = b[p]; };
    return function (d, b) {
        extendStatics(d, b);
        function __() { this.constructor = d; }
        d.prototype = b === null ? Object.create(b) : (__.prototype = b.prototype, new __());
    };
})();
Object.defineProperty(exports, "__esModule", { value: true });
// @ts-ignore
var promise = require("Promise");
if (promise && promise.polyfill) {
    promise.polyfill();
}
var AnimationCurve;
(function (AnimationCurve) {
    AnimationCurve["Linear"] = "linear";
    AnimationCurve["EaseIn"] = "easeIn";
    AnimationCurve["EaseOut"] = "easeOut";
    AnimationCurve["EaseInOut"] = "easeInOut";
})(AnimationCurve = exports.AnimationCurve || (exports.AnimationCurve = {}));
function _computeRender(elements, previousElements) {
    var out = {};
    for (var key in elements) {
        var element = elements[key];
        var hasNewAttribute = false;
        var newElement = null;
        var previousElement = previousElements ? previousElements[key] : undefined;
        if (element.children) {
            var children = element.children;
            var previousChildren = previousElement ? previousElement.children || [] : [];
            var newArray = [];
            for (var index = 0, max = children.length; index < max; index++) {
                var innerElements = children[index];
                var oldElements = index < previousChildren.length ? previousChildren[index] : undefined;
                var updatedElement = _computeRender(innerElements, oldElements);
                newArray.push(updatedElement);
            }
            hasNewAttribute = true;
            newElement = newArray;
        }
        else {
            var previousElement_1 = previousElements ? previousElements[key] : {};
            var elementAttributes = element.attributes || {};
            var previousElementAttributes = previousElement_1 ? previousElement_1.attributes : undefined;
            var newAttributes = {};
            for (var attribute in elementAttributes) {
                var attributeValue = elementAttributes[attribute];
                if (!previousElementAttributes || attributeValue !== previousElementAttributes[attribute]) {
                    if (attribute == 'viewModel' && runtime.makeOpaque) {
                        newAttributes[attribute] = runtime.makeOpaque(attributeValue);
                    }
                    else {
                        newAttributes[attribute] = attributeValue;
                    }
                    hasNewAttribute = true;
                }
            }
            newElement = newAttributes;
        }
        if (hasNewAttribute) {
            out[key] = newElement;
        }
    }
    return out;
}
function _copyElements(elements) {
    var outElements = {};
    for (var id in elements) {
        var outElement = {};
        var element = elements[id];
        if (element.attributes) {
            var elementAttributes = {};
            for (var attribute in element.attributes) {
                elementAttributes[attribute] = element.attributes[attribute];
            }
            outElement.attributes = elementAttributes;
        }
        if (element.children) {
            var outChildren = [];
            for (var _i = 0, _a = element.children; _i < _a.length; _i++) {
                var child = _a[_i];
                outChildren.push(_copyElements(child));
            }
            outElement.children = outChildren;
        }
        outElements[id] = outElement;
    }
    return outElements;
}
var componentById = {};
var Component = /** @class */ (function () {
    function Component(componentId, elements, actions, renderTemplateFunction, parentComponentId, elementIdWithinParent) {
        this.componentId = componentId;
        this.elements = elements;
        this.actions = actions;
        this.viewModel = null;
        this.parentComponentId = parentComponentId;
        this.elementIdWithinParent = elementIdWithinParent;
        this.children = {};
        if (renderTemplateFunction) {
            this.renderTemplateFunction = renderTemplateFunction.bind(this);
        }
    }
    Component.prototype.__doRender = function (animationOptions, renderTemplate, animationCompletionCallback) {
        if (renderTemplate && this.renderTemplateFunction) {
            this.renderTemplateFunction();
        }
        var elementsToRender = _computeRender(this.elements, this.previousElements);
        if (animationOptions != null) {
            runtime.postMessage(this.componentId, 'animate', {
                elements: elementsToRender,
                options: animationOptions,
                completionCallback: animationCompletionCallback,
            });
        }
        else {
            runtime.postMessage(this.componentId, 'render', elementsToRender);
        }
        this.previousElements = _copyElements(this.elements);
    };
    /**
     * Render which forces all attributes to be sent down to the runtime.
     */
    Component.prototype.forceRerender = function () {
        this.previousElements = undefined;
        this.render();
    };
    /**
     * Render the template (resolve all dynamic attributes in the template)
     * and send all changes to the runtime.
     */
    Component.prototype.render = function () {
        this.__doRender(null, true, null);
    };
    /**
     * Send all changes to the runtime without rendering the template (dynamic attributes
     * are not resolved again). This is useful if you are modifying some attributes directly
     * and only want to submit those.
     */
    Component.prototype.renderWithoutTemplate = function () {
        this.__doRender(null, false, null);
    };
    /**
     * Returns the view frames for the given elements.
     * @param elementIds
     * @param callback
     */
    Component.prototype.getElementFrames = function (elementIds, callback) {
        runtime.getFramesForElementIds(this.componentId, elementIds, callback);
    };
    /**
     * Animate the view tree using the given animation options.
     * The animations closure will be called
     */
    Component.prototype.animate = function (options, animations) {
        var _this = this;
        // Flush any pending attributes
        this.render();
        animations();
        return new Promise(function (resolve) {
            _this.__doRender(options, true, function () {
                resolve(undefined);
            });
        });
    };
    Component.prototype.onViewModelUpdated = function () {
    };
    Component.prototype.doOnChildCreated = function (childComponent) {
        var id = childComponent.elementIdWithinParent;
        var entries = this.children[id];
        if (entries === undefined) {
            entries = [];
        }
        entries.push(childComponent);
        this.children[id] = entries;
        this.onChildComponentCreated(childComponent);
    };
    Component.prototype.doOnChildComponentDestroyed = function (childComponent) {
        var id = childComponent.elementIdWithinParent;
        var entries = this.children[id];
        if (entries !== undefined) {
            var index = entries.indexOf(childComponent);
            if (index != -1) {
                entries.splice(index, 1);
            }
        }
        this.children[id] = entries;
        this.onChildComponentDestroyed(childComponent);
    };
    Component.prototype.doOnCreate = function () {
        if (this.parentComponentId) {
            this.parentComponent = Component.getComponent(this.parentComponentId);
        }
        this.onCreate();
        if (this.parentComponent) {
            this.parentComponent.doOnChildCreated(this);
        }
    };
    Component.prototype.doOnDestroy = function () {
        this.onDestroy();
        if (this.parentComponent) {
            this.parentComponent.doOnChildComponentDestroyed(this);
            this.parentComponent = null;
        }
        this.children = undefined;
        this.previousElements = undefined;
        this.elements = undefined;
    };
    /**
     * Called when a child Component was created, after its onCreate() has been called.
     */
    Component.prototype.onChildComponentCreated = function (childComponent) {
    };
    /**
     * Called when a child Component was created, after its onDestroy() has been called.
     */
    Component.prototype.onChildComponentDestroyed = function (childComponent) {
    };
    /**
     * Called when the component is created.
     */
    Component.prototype.onCreate = function () {
    };
    /**
     * Called wen the component was destroyed
     */
    Component.prototype.onDestroy = function () {
    };
    /**
     * Called whenever the layout has completed.
     */
    Component.prototype.onLayout = function () {
    };
    Component.getComponent = function (id) {
        return componentById[id];
    };
    Component.setComponent = function (id, component) {
        if (!component) {
            delete componentById[id];
        }
        else {
            componentById[id] = component;
        }
    };
    Component.initializeComponent = function (component) {
        Component.setComponent(component.componentId, component);
        component.doOnCreate();
    };
    Component.destroyComponent = function (componentId) {
        var component = Component.getComponent(componentId);
        if (component) {
            Component.setComponent(componentId, null);
            component.doOnDestroy();
        }
    };
    Component.updateViewModel = function (componentId, viewModel) {
        var component = Component.getComponent(componentId);
        if (component) {
            component.viewModel = viewModel;
            component.onViewModelUpdated();
            component.render();
        }
    };
    // DEPRECATED
    Component.getView = function (id) {
        return this.getComponent(id);
    };
    // DEPRECATED
    Component.setView = function (id, view) {
        this.setComponent(id, view);
    };
    return Component;
}());
exports.Component = Component;
/**
 * @deprecated, Use Component directly.
 */
var View = /** @class */ (function (_super) {
    __extends(View, _super);
    function View() {
        return _super !== null && _super.apply(this, arguments) || this;
    }
    return View;
}(Component));
exports.View = View;
function callInstanceFunction(id, functionName, parameters) {
    var component = Component.getComponent(id);
    if (!component) {
        return undefined;
    }
    var func = component[functionName];
    if (func === undefined) {
        throw new Error("Function " + functionName + " does not exist on component " + id);
    }
    return func.apply(component, parameters);
}
exports.callInstanceFunction = callInstanceFunction;
